"""
Configure the user's system for atopile development.
"""

from pathlib import Path
from textwrap import dedent

import rich
import rich.prompt

import atopile.version
import click

CONFIGURED_FOR_PATH = Path("~/.atopile/configured_for").expanduser().absolute()


def get_configured_for_version() -> atopile.version.Version:
    """Return the version of atopile that the user's system is configured for."""
    try:
        with CONFIGURED_FOR_PATH.open("r", encoding="utf-8") as f:
            version_str = f.read().strip()
    except FileNotFoundError:
        version_str = "0.0.0"
    return atopile.version.clean_version(atopile.version.Version.parse(version_str))


@click.command("configure")
def configure() -> None:
    """
    Configure the user's system for atopile development.
    """
    do_configure()


def do_configure_if_needed() -> None:
    """Configure the user's system for atopile development if it's not already configured."""
    if not CONFIGURED_FOR_PATH.parent.exists():
        rich.print(dedent(
            """
            Welcome! :partying_face:

            Looks like you're new to atopile, there's some initial setup we need to do.

            The following changes will be made:
            - Install the atopile KiCAD plugin for you
            """
        ))
        if rich.prompt.Confirm.ask(
            ":wrench: That cool :sunglasses:?", default="y"
        ):
            do_configure()
            return
        else:
            rich.print(":thumbs_up: No worries, you can always run `ato configure` later!")
            CONFIGURED_FOR_PATH.parent.mkdir(exist_ok=True, parents=True)
            return

    if not CONFIGURED_FOR_PATH.exists():
        # In this case the user has opted for not installing things
        return

    # Otherwise we're configured, but we might need to update
    do_configure()


def do_configure() -> None:
    """Perform system configuration required for atopile."""
    if get_configured_for_version() == atopile.version.get_installed_atopile_version():
        return

    CONFIGURED_FOR_PATH.parent.mkdir(parents=True, exist_ok=True)
    with CONFIGURED_FOR_PATH.open("w", encoding="utf-8") as f:
        f.write(
            str(
                atopile.version.clean_version(
                    atopile.version.get_installed_atopile_version()
                )
            )
        )

    # Otherwise, figure it out

    # FIXME: no idea what's up with this - but seem to help on Windows
    try:
        install_kicad_plugin()
    except FileNotFoundError:
        install_kicad_plugin()


PLUGIN_LOADER = f"""
plugin_path = "{Path(__file__).parent.parent}"
import sys
import importlib

if plugin_path not in sys.path:
    sys.path.append(plugin_path)

# if kicad_plugin is already in sys.modules, reload it
for module in sys.modules:
    if "kicad_plugin" in module:
        importlib.reload(sys.modules[module])

import kicad_plugin
"""


def install_kicad_plugin() -> None:
    """Install the kicad plugin."""
    # Find the path to kicad's plugin directory
    kicad_plugin_dir = (
        Path("~/Documents/KiCad/7.0/scripting/plugins").expanduser().absolute()
    )

    # Create the directory if it doesn't exist
    kicad_plugin_dir.mkdir(parents=True, exist_ok=True)

    # Write the plugin loader
    with (kicad_plugin_dir / "atopile.py").open("w", encoding="utf-8") as f:
        f.write(PLUGIN_LOADER)



# Generated from AtopileParser.g4 by ANTLR 4.13.0
from antlr4 import *
if "." in __name__:
    from .AtopileParser import AtopileParser
else:
    from AtopileParser import AtopileParser

# This class defines a complete listener for a parse tree produced by AtopileParser.
class AtopileParserListener(ParseTreeListener):

    # Enter a parse tree produced by AtopileParser#file_input.
    def enterFile_input(self, ctx:AtopileParser.File_inputContext):
        pass

    # Exit a parse tree produced by AtopileParser#file_input.
    def exitFile_input(self, ctx:AtopileParser.File_inputContext):
        pass


    # Enter a parse tree produced by AtopileParser#stmt.
    def enterStmt(self, ctx:AtopileParser.StmtContext):
        pass

    # Exit a parse tree produced by AtopileParser#stmt.
    def exitStmt(self, ctx:AtopileParser.StmtContext):
        pass


    # Enter a parse tree produced by AtopileParser#simple_stmts.
    def enterSimple_stmts(self, ctx:AtopileParser.Simple_stmtsContext):
        pass

    # Exit a parse tree produced by AtopileParser#simple_stmts.
    def exitSimple_stmts(self, ctx:AtopileParser.Simple_stmtsContext):
        pass


    # Enter a parse tree produced by AtopileParser#simple_stmt.
    def enterSimple_stmt(self, ctx:AtopileParser.Simple_stmtContext):
        pass

    # Exit a parse tree produced by AtopileParser#simple_stmt.
    def exitSimple_stmt(self, ctx:AtopileParser.Simple_stmtContext):
        pass


    # Enter a parse tree produced by AtopileParser#compound_stmt.
    def enterCompound_stmt(self, ctx:AtopileParser.Compound_stmtContext):
        pass

    # Exit a parse tree produced by AtopileParser#compound_stmt.
    def exitCompound_stmt(self, ctx:AtopileParser.Compound_stmtContext):
        pass


    # Enter a parse tree produced by AtopileParser#block.
    def enterBlock(self, ctx:AtopileParser.BlockContext):
        pass

    # Exit a parse tree produced by AtopileParser#block.
    def exitBlock(self, ctx:AtopileParser.BlockContext):
        pass


    # Enter a parse tree produced by AtopileParser#blockdef.
    def enterBlockdef(self, ctx:AtopileParser.BlockdefContext):
        pass

    # Exit a parse tree produced by AtopileParser#blockdef.
    def exitBlockdef(self, ctx:AtopileParser.BlockdefContext):
        pass


    # Enter a parse tree produced by AtopileParser#blocktype.
    def enterBlocktype(self, ctx:AtopileParser.BlocktypeContext):
        pass

    # Exit a parse tree produced by AtopileParser#blocktype.
    def exitBlocktype(self, ctx:AtopileParser.BlocktypeContext):
        pass


    # Enter a parse tree produced by AtopileParser#import_stmt.
    def enterImport_stmt(self, ctx:AtopileParser.Import_stmtContext):
        pass

    # Exit a parse tree produced by AtopileParser#import_stmt.
    def exitImport_stmt(self, ctx:AtopileParser.Import_stmtContext):
        pass


    # Enter a parse tree produced by AtopileParser#assign_stmt.
    def enterAssign_stmt(self, ctx:AtopileParser.Assign_stmtContext):
        pass

    # Exit a parse tree produced by AtopileParser#assign_stmt.
    def exitAssign_stmt(self, ctx:AtopileParser.Assign_stmtContext):
        pass


    # Enter a parse tree produced by AtopileParser#assignable.
    def enterAssignable(self, ctx:AtopileParser.AssignableContext):
        pass

    # Exit a parse tree produced by AtopileParser#assignable.
    def exitAssignable(self, ctx:AtopileParser.AssignableContext):
        pass


    # Enter a parse tree produced by AtopileParser#retype_stmt.
    def enterRetype_stmt(self, ctx:AtopileParser.Retype_stmtContext):
        pass

    # Exit a parse tree produced by AtopileParser#retype_stmt.
    def exitRetype_stmt(self, ctx:AtopileParser.Retype_stmtContext):
        pass


    # Enter a parse tree produced by AtopileParser#connect_stmt.
    def enterConnect_stmt(self, ctx:AtopileParser.Connect_stmtContext):
        pass

    # Exit a parse tree produced by AtopileParser#connect_stmt.
    def exitConnect_stmt(self, ctx:AtopileParser.Connect_stmtContext):
        pass


    # Enter a parse tree produced by AtopileParser#connectable.
    def enterConnectable(self, ctx:AtopileParser.ConnectableContext):
        pass

    # Exit a parse tree produced by AtopileParser#connectable.
    def exitConnectable(self, ctx:AtopileParser.ConnectableContext):
        pass


    # Enter a parse tree produced by AtopileParser#signaldef_stmt.
    def enterSignaldef_stmt(self, ctx:AtopileParser.Signaldef_stmtContext):
        pass

    # Exit a parse tree produced by AtopileParser#signaldef_stmt.
    def exitSignaldef_stmt(self, ctx:AtopileParser.Signaldef_stmtContext):
        pass


    # Enter a parse tree produced by AtopileParser#pindef_stmt.
    def enterPindef_stmt(self, ctx:AtopileParser.Pindef_stmtContext):
        pass

    # Exit a parse tree produced by AtopileParser#pindef_stmt.
    def exitPindef_stmt(self, ctx:AtopileParser.Pindef_stmtContext):
        pass


    # Enter a parse tree produced by AtopileParser#with_stmt.
    def enterWith_stmt(self, ctx:AtopileParser.With_stmtContext):
        pass

    # Exit a parse tree produced by AtopileParser#with_stmt.
    def exitWith_stmt(self, ctx:AtopileParser.With_stmtContext):
        pass


    # Enter a parse tree produced by AtopileParser#new_stmt.
    def enterNew_stmt(self, ctx:AtopileParser.New_stmtContext):
        pass

    # Exit a parse tree produced by AtopileParser#new_stmt.
    def exitNew_stmt(self, ctx:AtopileParser.New_stmtContext):
        pass


    # Enter a parse tree produced by AtopileParser#name_or_attr.
    def enterName_or_attr(self, ctx:AtopileParser.Name_or_attrContext):
        pass

    # Exit a parse tree produced by AtopileParser#name_or_attr.
    def exitName_or_attr(self, ctx:AtopileParser.Name_or_attrContext):
        pass


    # Enter a parse tree produced by AtopileParser#numerical_pin_ref.
    def enterNumerical_pin_ref(self, ctx:AtopileParser.Numerical_pin_refContext):
        pass

    # Exit a parse tree produced by AtopileParser#numerical_pin_ref.
    def exitNumerical_pin_ref(self, ctx:AtopileParser.Numerical_pin_refContext):
        pass


    # Enter a parse tree produced by AtopileParser#attr.
    def enterAttr(self, ctx:AtopileParser.AttrContext):
        pass

    # Exit a parse tree produced by AtopileParser#attr.
    def exitAttr(self, ctx:AtopileParser.AttrContext):
        pass


    # Enter a parse tree produced by AtopileParser#totally_an_integer.
    def enterTotally_an_integer(self, ctx:AtopileParser.Totally_an_integerContext):
        pass

    # Exit a parse tree produced by AtopileParser#totally_an_integer.
    def exitTotally_an_integer(self, ctx:AtopileParser.Totally_an_integerContext):
        pass


    # Enter a parse tree produced by AtopileParser#name.
    def enterName(self, ctx:AtopileParser.NameContext):
        pass

    # Exit a parse tree produced by AtopileParser#name.
    def exitName(self, ctx:AtopileParser.NameContext):
        pass


    # Enter a parse tree produced by AtopileParser#string.
    def enterString(self, ctx:AtopileParser.StringContext):
        pass

    # Exit a parse tree produced by AtopileParser#string.
    def exitString(self, ctx:AtopileParser.StringContext):
        pass


    # Enter a parse tree produced by AtopileParser#boolean_.
    def enterBoolean_(self, ctx:AtopileParser.Boolean_Context):
        pass

    # Exit a parse tree produced by AtopileParser#boolean_.
    def exitBoolean_(self, ctx:AtopileParser.Boolean_Context):
        pass



del AtopileParser


from atopile import config


def test_sanitise_dict_keys():
    """Test that dict keys are sanitised."""
    assert config._sanitise_dict_keys({"a-b": 1, "ato-version": {"e-f": 2}}) == {
        "a-b": 1,
        "ato_version": {"e-f": 2},
    }



import pytest

from atopile.errors import AtoError, _error_accumulator, iter_through_errors


def test_error_accumulator():
    with pytest.raises(ExceptionGroup):
        with _error_accumulator() as err_cltr:
            with err_cltr():
                raise AtoError("test error")

            # FIXME: damn... I don't like that the type-checker/linter
            # doesn't realise the error is supressed
            with err_cltr():
                raise AtoError("test error 2")


def test_iter_through_errors():
    try:
        for cltr, i in iter_through_errors(range(4)):
            with cltr():
                if i == 1:
                    raise AtoError("test error")
                if i == 2:
                    raise AtoError("test error 2")

    except ExceptionGroup as ex:
        assert len(ex.exceptions) == 2
        ex_1, ex_2 = ex.exceptions
        assert ex_1.message == "test error"
        assert ex_2.message == "test error 2"

    else:
        raise AssertionError("Expected an ExceptionGroup to be raised")



# Copyright (c) 2016-2024 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import annotations

import posixpath
import re

from mkdocs.config.defaults import MkDocsConfig
from mkdocs.structure.files import File, Files
from mkdocs.structure.pages import Page
from re import Match

# -----------------------------------------------------------------------------
# Hooks
# -----------------------------------------------------------------------------

# @todo
def on_page_markdown(
    markdown: str, *, page: Page, config: MkDocsConfig, files: Files
):

    # Replace callback
    def replace(match: Match):
        type, args = match.groups()
        args = args.strip()
        if type == "version":
            if args.startswith("insiders-"):
                return _badge_for_version_insiders(args, page, files)
            else:
                return _badge_for_version(args, page, files)
        elif type == "sponsors":     return _badge_for_sponsors(page, files)
        elif type == "flag":         return flag(args, page, files)
        elif type == "option":       return option(args)
        elif type == "setting":      return setting(args)
        elif type == "feature":      return _badge_for_feature(args, page, files)
        elif type == "plugin":       return _badge_for_plugin(args, page, files)
        elif type == "extension":    return _badge_for_extension(args, page, files)
        elif type == "utility":      return _badge_for_utility(args, page, files)
        elif type == "example":      return _badge_for_example(args, page, files)
        elif type == "default":
            if   args == "none":     return _badge_for_default_none(page, files)
            elif args == "computed": return _badge_for_default_computed(page, files)
            else:                    return _badge_for_default(args, page, files)

        # Otherwise, raise an error
        raise RuntimeError(f"Unknown shortcode: {type}")

    # Find and replace all external asset URLs in current page
    return re.sub(
        r"<!-- md:(\w+)(.*?) -->",
        replace, markdown, flags = re.I | re.M
    )

# -----------------------------------------------------------------------------
# Helper functions
# -----------------------------------------------------------------------------

# Create a flag of a specific type
def flag(args: str, page: Page, files: Files):
    type, *_ = args.split(" ", 1)
    if   type == "experimental":  return _badge_for_experimental(page, files)
    elif type == "required":      return _badge_for_required(page, files)
    elif type == "customization": return _badge_for_customization(page, files)
    elif type == "metadata":      return _badge_for_metadata(page, files)
    elif type == "multiple":      return _badge_for_multiple(page, files)
    raise RuntimeError(f"Unknown type: {type}")

# Create a linkable option
def option(type: str):
    _, *_, name = re.split(r"[.:]", type)
    return f"[`{name}`](#+{type}){{ #+{type} }}\n\n"

# Create a linkable setting - @todo append them to the bottom of the page
def setting(type: str):
    _, *_, name = re.split(r"[.*]", type)
    return f"`{name}` {{ #{type} }}\n\n[{type}]: #{type}\n\n"

# -----------------------------------------------------------------------------

# Resolve path of file relative to given page - the posixpath always includes
# one additional level of `..` which we need to remove
def _resolve_path(path: str, page: Page, files: Files):
    path, anchor, *_ = f"{path}#".split("#")
    path = _resolve(files.get_file_from_path(path), page)
    return "#".join([path, anchor]) if anchor else path

# Resolve path of file relative to given page - the posixpath always includes
# one additional level of `..` which we need to remove
def _resolve(file: File, page: Page):
    path = posixpath.relpath(file.src_uri, page.file.src_uri)
    return posixpath.sep.join(path.split(posixpath.sep)[1:])

# -----------------------------------------------------------------------------

# Create badge
def _badge(icon: str, text: str = "", type: str = ""):
    classes = f"mdx-badge mdx-badge--{type}" if type else "mdx-badge"
    text = f"{text}{{ data-preview='' }}" if text.endswith(")") else text
    return "".join([
        f"<span class=\"{classes}\">",
        *([f"<span class=\"mdx-badge__icon\">{icon}</span>"] if icon else []),
        *([f"<span class=\"mdx-badge__text\">{text}</span>"] if text else []),
        "</span>",
    ])

# Create sponsors badge
def _badge_for_sponsors(page: Page, files: Files):
    icon = "material-heart"
    href = _resolve_path("insiders/index.md", page, files)
    return _badge(
        icon = f"[:{icon}:]({href} 'Sponsors only')",
        type = "heart"
    )

# Create badge for version
def _badge_for_version(text: str, page: Page, files: Files):
    spec = text
    path = f"changelog/index.md#{spec}"

    # Return badge
    icon = "material-tag-outline"
    href = _resolve_path("conventions.md#version", page, files)
    return _badge(
        icon = f"[:{icon}:]({href} 'Minimum version')",
        text = f"[{text}]({_resolve_path(path, page, files)})" if spec else ""
    )

# Create badge for version of Insiders
def _badge_for_version_insiders(text: str, page: Page, files: Files):
    spec = text.replace("insiders-", "")
    path = f"insiders/changelog/index.md#{spec}"

    # Return badge
    icon = "material-tag-heart-outline"
    href = _resolve_path("conventions.md#version-insiders", page, files)
    return _badge(
        icon = f"[:{icon}:]({href} 'Minimum version')",
        text = f"[{text}]({_resolve_path(path, page, files)})" if spec else ""
    )

# Create badge for feature
def _badge_for_feature(text: str, page: Page, files: Files):
    icon = "material-toggle-switch"
    href = _resolve_path("conventions.md#feature", page, files)
    return _badge(
        icon = f"[:{icon}:]({href} 'Optional feature')",
        text = text
    )

# Create badge for plugin
def _badge_for_plugin(text: str, page: Page, files: Files):
    icon = "material-floppy"
    href = _resolve_path("conventions.md#plugin", page, files)
    return _badge(
        icon = f"[:{icon}:]({href} 'Plugin')",
        text = text
    )

# Create badge for extension
def _badge_for_extension(text: str, page: Page, files: Files):
    icon = "material-language-markdown"
    href = _resolve_path("conventions.md#extension", page, files)
    return _badge(
        icon = f"[:{icon}:]({href} 'Markdown extension')",
        text = text
    )

# Create badge for utility
def _badge_for_utility(text: str, page: Page, files: Files):
    icon = "material-package-variant"
    href = _resolve_path("conventions.md#utility", page, files)
    return _badge(
        icon = f"[:{icon}:]({href} 'Third-party utility')",
        text = text
    )

# Create badge for example
def _badge_for_example(text: str, page: Page, files: Files):
    return "\n".join([
        _badge_for_example_download(text, page, files),
        _badge_for_example_view(text, page, files)
    ])

# Create badge for example view
def _badge_for_example_view(text: str, page: Page, files: Files):
    icon = "material-folder-eye"
    href = f"https://mkdocs-material.github.io/examples/{text}/"
    return _badge(
        icon = f"[:{icon}:]({href} 'View example')",
        type = "right"
    )

# Create badge for example download
def _badge_for_example_download(text: str, page: Page, files: Files):
    icon = "material-folder-download"
    href = f"https://mkdocs-material.github.io/examples/{text}.zip"
    return _badge(
        icon = f"[:{icon}:]({href} 'Download example')",
        text = f"[`.zip`]({href})",
        type = "right"
    )

# Create badge for default value
def _badge_for_default(text: str, page: Page, files: Files):
    icon = "material-water"
    href = _resolve_path("conventions.md#default", page, files)
    return _badge(
        icon = f"[:{icon}:]({href} 'Default value')",
        text = text
    )

# Create badge for empty default value
def _badge_for_default_none(page: Page, files: Files):
    icon = "material-water-outline"
    href = _resolve_path("conventions.md#default", page, files)
    return _badge(
        icon = f"[:{icon}:]({href} 'Default value is empty')"
    )

# Create badge for computed default value
def _badge_for_default_computed(page: Page, files: Files):
    icon = "material-water-check"
    href = _resolve_path("conventions.md#default", page, files)
    return _badge(
        icon = f"[:{icon}:]({href} 'Default value is computed')"
    )

# Create badge for metadata property flag
def _badge_for_metadata(page: Page, files: Files):
    icon = "material-list-box-outline"
    href = _resolve_path("conventions.md#metadata", page, files)
    return _badge(
        icon = f"[:{icon}:]({href} 'Metadata property')"
    )

# Create badge for required value flag
def _badge_for_required(page: Page, files: Files):
    icon = "material-alert"
    href = _resolve_path("conventions.md#required", page, files)
    return _badge(
        icon = f"[:{icon}:]({href} 'Required value')"
    )

# Create badge for customization flag
def _badge_for_customization(page: Page, files: Files):
    icon = "material-brush-variant"
    href = _resolve_path("conventions.md#customization", page, files)
    return _badge(
        icon = f"[:{icon}:]({href} 'Customization')"
    )

# Create badge for multiple instance flag
def _badge_for_multiple(page: Page, files: Files):
    icon = "material-inbox-multiple"
    href = _resolve_path("conventions.md#multiple-instances", page, files)
    return _badge(
        icon = f"[:{icon}:]({href} 'Multiple instances')"
    )

# Create badge for experimental flag
def _badge_for_experimental(page: Page, files: Files):
    icon = "material-flask-outline"
    href = _resolve_path("conventions.md#experimental", page, files)
    return _badge(
        icon = f"[:{icon}:]({href} 'Experimental')"
    )



"""
Generate a report based on assertions made in the source code.
"""

import logging
from enum import Enum
from numbers import Number

import rich
from rich.style import Style
from rich.table import Table

from atopile import address, config, errors
from atopile.front_end import Assertion, Instance, RangedValue, lofty
from atopile.instance_methods import all_descendants

log = logging.getLogger(__name__)

light_row = Style(color="bright_black")
dark_row = Style(color="white")


class AssertionStatus(Enum):
    """
    The status of an assertion.
    """

    PASSED = "[green]Passed[/]"
    FAILED = "[red]Failed[/]"
    ERROR = "[red]Error[/]"


def generate_assertion_report(build_ctx: config.BuildContext):
    """
    Generate a report based on assertions made in the source code.
    """
    table = Table(show_header=True, header_style="bold green")
    table.add_column("Status")
    table.add_column("Assertion")
    table.add_column("Numeric")
    table.add_column("Address")
    # table.add_column("Notes")

    rows = 0

    def _add_row(
        status: AssertionStatus,
        assertion_str: str,
        numeric: str,
        addr: address.AddrStr,
        notes: str,
    ):
        nonlocal rows
        table.add_row(
            status.value,
            assertion_str,
            numeric,
            address.get_instance_section(addr),
            # notes,  # TODO: re-add notes
            style=dark_row if rows % 2 else light_row,
        )
        rows += 1

    def _check_assertion(instance: Instance, assertion: Assertion) -> AssertionStatus:
        """
        Check an assertion.
        """
        stmt = assertion.assertion_str

        ops = _which_operators(stmt)
        if len(ops) != 1:
            raise ValueError(
                f"Statement must contain exactly one of the following operators: {OPERATORS}"
            )

        op = ops[0]
        expr_str, bounds_str = stmt.split(op)

        context = _get_context(instance)

        try:
            # This is split in two due to the "within" operator
            expr = eval(expr_str, context)
            bounds = eval(bounds_str, context)
        except (NameError, AttributeError) as ex:
            raise errors.AtoKeyError(f"NameError: {ex}") from ex
            # Hacking some mildly better exceptions since the
            # parser currently doesn't actually, well, parse
            # the expressions in assertions.
            # dotted_path = ex.name.split(".")

            # TODO:
            # 1. Check if something's declared, but not defined
            # 2. If it's not defined or declared, at least convert
            # it to an ato error and re-raise as an ato error
            # if len(dotted_path) == 1:
            #     final_instance = instance
            # try:
            #     final_instance = _follow_the_dots(context, dotted_path[:-1])
            # except AttributeError:
            #     raise

        if _do_op(op, expr, bounds):
            status = AssertionStatus.PASSED
        else:
            status = AssertionStatus.FAILED

        numeric = str(expr) + " " + op + " " + str(bounds)

        _add_row(status, assertion.assertion_str, numeric, instance_addr, "")

    for instance_addr in all_descendants(build_ctx.entry):
        instance = lofty.get_instance(instance_addr)
        for assertion in instance.assertions:
            # try:
            _check_assertion(instance, assertion)
            # except Exception as ex:
            #     log.error(  # pylint: disable=logging-fstring-interpolation
            #         f"Error checking assertion '{assertion}' at {instance_addr}: {ex}"
            #     )
            #     _add_row(AssertionStatus.ERROR, assertion, instance_addr, str(ex))

    rich.print(table)


OPERATORS = ["<", "<=", ">", ">=", "within"]


def _which_operators(stmt: str) -> list[str]:
    """Return which operator is being used by an assertion."""
    return [op for op in OPERATORS if op in stmt]


def _do_op(op: str, this: RangedValue, other: RangedValue) -> bool:
    """Perform the operation specified by the operator."""
    if op == "within":
        return this.within(other)
    elif op == "<":
        return this < other
    elif op == "<=":
        return this <= other
    elif op == ">":
        return this > other
    elif op == ">=":
        return this >= other
    else:
        raise ValueError(f"Unrecognized operator: {op}")


class DotDict(dict):
    """A dict you can dot"""

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as ex:
            raise AttributeError(f"Attribute {name} not found", name) from ex


def _get_context(instance: Instance) -> dict:
    """
    Get the context for evaluating assertions.
    TODO: wtf is this? come on Matt
    """
    context = {}
    # Add assignments' values
    for k, a in instance.assignments.items():
        v = a[0].value
        if isinstance(v, (Number, RangedValue)):
            context[k] = v

    # Recurse through children
    for k, v in instance.children.items():
        context[k] = DotDict(_get_context(v))

    return context



# pylint: disable=logging-fstring-interpolation

"""
`ato inspect`
"""

import itertools
import logging
from typing import Optional

import click
import natsort
import rich
from rich.table import Table

from atopile import address, errors
from atopile.address import AddrStr
from atopile.cli.common import project_options
from atopile.config import BuildContext
from atopile.front_end import Link, lofty
from atopile.instance_methods import (
    all_descendants,
    get_links,
    get_parent,
    iter_parents,
    match_interfaces,
    match_modules,
    match_pins,
    match_pins_and_signals,
    match_signals,
)

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

class DisplayEntry:
    def __init__(self, connectables: list[AddrStr]):
        self.connectables: list[AddrStr] = connectables
        self.intermediate_connectable: list[AddrStr] = []
        self.consumers: list[AddrStr] = []


class InspectConnectable:
    def __init__(self, addr: AddrStr, type: str):
        self.addr = addr
        self.type = type
        self.associated_connectable = []
        self.parent_interface = None
        self.consumer = []


def collect_connected_dfs(connections: list[Link], root_addr, visited=None) -> list[AddrStr]:
    """
    from a list of links and a root node, this function returns all the nodes connected to the root node
    """
    if visited is None:
        visited = []

    for connection in connections:
        if connection.source.addr == root_addr:
            if connection.target.addr not in visited:
                visited.append(connection.target.addr)  # Mark this target as visited
                # Recursively visit the targets connected to this target
                collect_connected_dfs(connections, connection.target.addr, visited)
        elif connection.target.addr == root_addr:
            if connection.source.addr not in visited:
                visited.append(connection.source.addr)  # Mark this target as visited
                # Recursively visit the targets connected to this target
                collect_connected_dfs(connections, connection.source.addr, visited)
    return visited

def find_nets(graph) -> list[list[AddrStr]]:
    """
    For a dict of connections:
    {
    A: [B, C],
    B: [A, C],
    C: [A, B],
    D: [E],
    E: [D]
    }
    This function returns the nets:
    [[A, B, C], [D, E]]
    """
    visited = set()  # Keep track of visited nodes to avoid revisiting them.
    connected_components = []  # Store the connected components.
    def dfs(node, component):
        """
        Perform depth-first search recursively to find all nodes connected to the current node.
        """
        visited.add(node)
        component.append(node)
        for neighbor in graph.get(node, []):
            if neighbor not in visited:
                dfs(neighbor, component)

    for node in graph:
        if node not in visited:
            component = []
            dfs(node, component)
            connected_components.append(component)

    return connected_components

odd_row = "on grey11 cornflower_blue"
even_row = "on grey15 cornflower_blue"
odd_greyed_row = "on grey11 grey0"
even_greyed_row = "on grey15 grey0"


@click.command()
@project_options
@click.option("--inspect", required=True)
@click.option("--context", default=None, help="The context from which to inspect the module")
@errors.muffle_fatalities
def inspect(build_ctxs: list[BuildContext], inspect: str, context: Optional[str]):
    """
    Utility to inspect what is connected to a component.
    The context set the boundary where something is considered connecting to it.
    For example: --inspect rp2040_micro --context rp2040_micro_kit
    """
    if len(build_ctxs) == 0:
        errors.AtoNotImplementedError("No build contexts found.")
    elif len(build_ctxs) == 1:
        build_ctx = build_ctxs[0]
    else:
        build_ctx = build_ctxs[0]
        errors.AtoNotImplementedError(
            f"Using top build config {build_ctx.name} for now. Multiple build configs not yet supported."
        ).log(log, logging.WARNING)

    #TODO: make sure that the context is always above the module to inspect
    module_to_inspect = address.add_instance(build_ctx.entry, inspect)
    if context is None:
        from_perspective_of_module = module_to_inspect
    else:
        from_perspective_of_module = address.add_instance(build_ctx.entry, context)

    log.info(f"Inspecting {address.get_instance_section(module_to_inspect)} from the perspective of {address.get_instance_section(from_perspective_of_module)}")

    # TODO: Currently doing this just to fill the cache
    lofty.get_instance(build_ctx.entry)

    # Find the interface modules
    inspect_parent_modules = list(iter_parents(module_to_inspect))
    context_child_modules = list(filter(match_modules, all_descendants(from_perspective_of_module)))
    interface_modules = list(set(inspect_parent_modules).intersection(set(context_child_modules)))

    # Create an array of all the connectables we could inspect
    inspected_connectables: dict[AddrStr, InspectConnectable] = {}
    for connectable in list(filter(match_pins, all_descendants(from_perspective_of_module))):
        inspected_connectables[connectable] = InspectConnectable(connectable, "pin")
    for connectable in list(filter(match_signals, all_descendants(from_perspective_of_module))):
        inspected_connectables[connectable] = InspectConnectable(connectable, "signal")
    for connectable in list(filter(match_interfaces, all_descendants(from_perspective_of_module))):
        inspected_connectables[connectable] = InspectConnectable(connectable, "interface")

    # If a signal has a parent interface, add it to the inspected object
    for conn in inspected_connectables:
        if inspected_connectables[conn].type == 'interface':
            signals_to_update = list(filter(match_signals, all_descendants(conn)))
            for signal in signals_to_update:
                inspected_connectables[signal].parent_interface = conn

    # Save the links that are in the current module
    context_modules = [from_perspective_of_module]
    context_modules.extend(list(filter(match_modules, all_descendants(from_perspective_of_module))))
    context_child_links: list[AddrStr] = []
    for module in context_modules:
        context_child_links.extend(list(get_links(module)))

    context_connection_graph = {}
    # For each connectable in the context, find connectables connected to it
    for connectable in inspected_connectables:
        context_connected_conns = collect_connected_dfs(context_child_links, connectable)
        context_connection_graph[connectable] = context_connected_conns
        inspected_connectables[connectable].associated_connectable = context_connected_conns

    context_to_inspect_nets: list[list[AddrStr]] = []
    context_to_inspect_nets = find_nets(context_connection_graph)

    # Save the links that are above the current module
    parents = list(iter_parents(from_perspective_of_module))
    parent_links = []
    for parent in parents:
        parent_links.extend(list(get_links(parent)))

    # For each connectable in the context, find what consumes it
    for connectable in inspected_connectables:
        for link in parent_links:
            if link.source.addr == connectable:
                inspected_connectables[connectable].consumer.append(link.target.addr)
            if link.target.addr == connectable:
                inspected_connectables[connectable].consumer.append(link.source.addr)

    ####
    # Inspected module nets
    ####
    inspect_connectables = list(filter(match_pins_and_signals, all_descendants(module_to_inspect)))
    inspect_connectables.extend(list(filter(match_interfaces, all_descendants(module_to_inspect))))

    # Create nets for the elements in the component to inspect
    module_connection_graph = {}
    module_to_inspect_child_links = list(get_links(module_to_inspect))
    for connectable in inspect_connectables:
        inspect_module_connected_conns = collect_connected_dfs(module_to_inspect_child_links, connectable)
        module_connection_graph[connectable] = inspect_module_connected_conns

    module_to_inspect_nets: list[list[AddrStr]] = []
    module_to_inspect_nets = find_nets(module_connection_graph)

    # Make a map between the connectables in the inspected module and the nets in the context
    module_conn_to_context_net_map: dict[AddrStr, list[AddrStr]] = {}
    for module_net in module_to_inspect_nets:
        for conn in module_net:
            for context_net in context_to_inspect_nets:
                if conn in context_net:
                    module_conn_to_context_net_map[conn] = context_net
                    break
                else:
                    module_conn_to_context_net_map[conn] = []

    # Create the display entries
    displayed_entries: list[DisplayEntry] = []
    # There is one entry per net
    for net in module_to_inspect_nets:
        disp_entry = DisplayEntry(net)
        # Add the intermediate interfaces
        # Add the consumers information
        for conn in net:
            for tested_conn in module_conn_to_context_net_map[conn]:
                if inspected_connectables[tested_conn].consumer != []:
                    disp_entry.consumers.extend(inspected_connectables[tested_conn].consumer)
                parent_interface = inspected_connectables[tested_conn].parent_interface
                if parent_interface is not None:
                    if inspected_connectables[parent_interface].consumer != []:
                        for consumer in inspected_connectables[parent_interface].consumer:
                            disp_entry.consumers.append(str(consumer + "." + address.get_name(tested_conn)))
                    if get_parent(parent_interface) in interface_modules:
                        disp_entry.intermediate_connectable.append(tested_conn)
                if get_parent(tested_conn) in interface_modules:
                    disp_entry.intermediate_connectable.append(tested_conn)

        displayed_entries.append(disp_entry)
        disp_entry.connectables.sort()

    # Create a table
    inspection_table = Table(show_header=True, header_style="bold cornflower_blue")
    inspection_table.add_column("Pin #", justify="right")
    inspection_table.add_column("Signal name", justify="left")
    inspection_table.add_column("Interface", justify="left")
    inspection_table.add_column("Connected to", justify="left")

    # Help to fill the table
    bom_row_nb_counter = itertools.count()
    def _add_row(pins, signals, intermediate, consumers):
        row_nb = next(bom_row_nb_counter)
        if consumers == []:
            inspection_table.add_row(
                f"{', '.join([address.get_name(x) for x in pins])}",
                f"{', '.join([address.get_name(x) for x in signals])}",
                f"{', '.join([address.get_instance_section(x) for x in intermediate])}",
                f"{', '.join([address.get_instance_section(x) for x in consumers])}",
                style=even_row if row_nb % 2 else odd_row,
            )
        else:
            inspection_table.add_row(
                f"{', '.join([address.get_name(x) for x in pins])}",
                f"{', '.join([address.get_name(x) for x in signals])}",
                f"{', '.join([address.get_instance_section(x) for x in intermediate])}",
                f"{', '.join([address.get_instance_section(x) for x in consumers])}",
                style=even_greyed_row if row_nb % 2 else odd_greyed_row,
            )

    sorted_entries = natsort.natsorted(displayed_entries, key=lambda obj: obj.connectables[0])
    for entry in sorted_entries:
        pins = sorted(list(filter(match_pins, entry.connectables)))
        signals = list(filter(match_signals, entry.connectables))
        _add_row(pins, signals, list(set(entry.intermediate_connectable)), list(set(entry.consumers)))


    rich.print(inspection_table)



import pytest

from atopile import assertions


def test_DotDict():
    d = assertions.DotDict({"a": 1, "b": 2})
    assert d.a == 1
    assert d.b == 2

    with pytest.raises(AttributeError):
        d.c


def test_follow_the_dots():
    d = assertions.DotDict({"a": {"b": {"c": 1}}})
    assert assertions._follow_the_dots(d, ["a", "b", "c"]) == 1



import logging
from pathlib import Path

import pcbnew

from .common import (
    get_layout_map,
    groups_by_name,
    get_footprint_uuid,
    footprints_by_uuid,
)

log = logging.getLogger(__name__)


class ReloadGroup(pcbnew.ActionPlugin):
    def defaults(self):
        self.name = "Reload Group"
        self.category = "Reload Group Layout Atopile"
        self.description = "Layout components on PCB in same spatial relationships as components on schematic"
        self.show_toolbar_button = True
        self.icon_file_name = str(Path(__file__).parent / "reload.png")
        self.dark_icon_file_name = self.icon_file_name

    def Run(self):
        board: pcbnew.BOARD = pcbnew.GetBoard()
        board_path = board.GetFileName()

        existing_groups = groups_by_name(board)
        footprints = footprints_by_uuid(board)

        for group_name, group_data in get_layout_map(board_path).items():
            # If the group doesn't yet exist in the layout
            # create it and add it to the board
            if group_name in existing_groups:
                g = existing_groups[group_name]
            else:
                g = pcbnew.PCB_GROUP(board)
                g.SetName(group_name)
                board.Add(g)

            # Make sure all the footprints in the group are up to date
            footprints_in_group = {
                get_footprint_uuid(fp)
                for fp in g.GetItems()
                if isinstance(fp, pcbnew.FOOTPRINT)
            }
            expected_footprints = set(group_data["uuid_map"].keys())
            for fp_uuid in footprints_in_group - expected_footprints:
                g.RemoveItem(footprints[fp_uuid])
            for fp_uuid in expected_footprints - footprints_in_group:
                g.AddItem(footprints[fp_uuid])


ReloadGroup().register()



from atopile.address import get_file, get_instance_section


def test_get_file():
    # Test with a file address
    assert get_file("path/to/file.txt") == "path/to/file.txt"

    # Test with a file address containing special characters
    assert (
        get_file("path/to/file-with-special-chars.txt")
        == "path/to/file-with-special-chars.txt"
    )

    # Test with a file address containing a Windows drive letter
    assert get_file("C:/path/to/file.txt") == "C:/path/to/file.txt"

    # Test with an empty file address
    assert get_file("") == ""

    assert get_file("path/to/file.txt:a:b.c") == "path/to/file.txt"

    # Test with a file address containing special characters
    assert (
        get_file("path/to/file-with-special-chars.txt:a:b.c")
        == "path/to/file-with-special-chars.txt"
    )

    # Test with a file address containing a Windows drive letter
    assert get_file("C:/path/to/file.txt:a:b.c") == "C:/path/to/file.txt"


def test_get_instance_section():
    # Test with an address containing an instance section
    assert get_instance_section("path/to/file.txt:a:b.c") is None

    # Test with an address not containing an instance section
    assert get_instance_section("path/to/file.txt") is None

    # Test with an empty address
    assert get_instance_section("") is None

    # Test with an address containing a Windows drive letter
    assert get_instance_section("C:/path/to/file.txt:a::b.c") == "b.c"

    # Test with an address containing a Windows drive letter
    assert get_instance_section("C:/path/to/file-with-special-chars.txt:a::b.c") == "b.c"
